/*
 * sudo_server.c
 */

#include <windows.h>
#include <io.h>
#include <fcntl.h>
#include <stdio.h>
#include <time.h>
#include "sudo_server.h"
#include "utils.h"

#define PARAM_SIZE  (9+NAME_SIZE+1+PATH_SIZE)

static HANDLE hCEvent;

/* OtpCv̎M */
static DWORD WINAPI pipeReader(LPVOID lpParam) {
	HANDLE hPipe = (HANDLE)lpParam;
	char szBuf[1024];
	DWORD dwRead;
	DWORD dwWritten;
	HANDLE hStdout;
	
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	
	while (1) {
		ConnectNamedPipe(hPipe, NULL);
		
		if (!ReadFile(hPipe, szBuf, sizeof(szBuf) - 1, &dwRead, NULL)) break;
		if (dwRead > 0) WriteFile(hStdout, szBuf, dwRead, &dwWritten, NULL);
	}
	
	return 0;
}

static int envlen(const char *env) {
	int len = 0;
	
	for (; env[len]; len++) {
		while (env[len]) len++;
	}
	
	return len + 1;
}

static DWORD WINAPI writeEnv(LPVOID lpParam) {
	HANDLE hEnvPipe = (HANDLE)lpParam;
	LPVOID env;
	DWORD dwWritten;
	
	env = GetEnvironmentStrings();
	
	ConnectNamedPipe(hEnvPipe, NULL);
	WriteFile(hEnvPipe, env, envlen(env), &dwWritten, NULL);
	
	FreeEnvironmentStrings(env);
	
	return 0;
}

/* Ctrl+C */
static BOOL WINAPI HandlerRoutine(DWORD dwCtrlType) {
	if (dwCtrlType == CTRL_C_EVENT || dwCtrlType == CTRL_CLOSE_EVENT) {
		PulseEvent(hCEvent);
		return TRUE;
	}
	
	return FALSE;
}

static char *getID() {
	static char pipeName[NAME_SIZE];
	sprintf(pipeName, "%ld_%ld", GetCurrentProcessId(), time(NULL));
	
	return pipeName;
}

/* M */
void serverMain(const char *sudo, const char *command, char debug_mode) {
	char lpParam[PARAM_SIZE];
	char *pipeName;
	HANDLE hPipe;
	HANDLE hEnvPipe;
	HANDLE hThread;
	LPCTSTR lpVerb;
	SHELLEXECUTEINFO si = {0};
	
	pipeName = getID();
	lpVerb = debug_mode ? "open" : "runas";
	
	if (9+strlen(pipeName)+1+strlen(command)+1 > PARAM_SIZE) exit(1);
	sprintf(lpParam, "--client %s %s", pipeName, command);
	
	/*I*/
	hCEvent = CreateEvent(NULL, FALSE, FALSE, getCEventName(pipeName));
	if (hCEvent == NULL) return;
	SetConsoleCtrlHandler(HandlerRoutine, TRUE);
	
	/*pCv쐬*/
	hPipe = CreateNamedPipe(getPipeName(pipeName),
		PIPE_ACCESS_INBOUND, PIPE_TYPE_BYTE,
		1,  4095, 4095, 0, NULL);
	if (hPipe == INVALID_HANDLE_VALUE) return;
	
	hEnvPipe = CreateNamedPipe(getEnvPipeName(pipeName),
		PIPE_ACCESS_OUTBOUND, PIPE_TYPE_BYTE,
		1,  4095, 4095, 0, NULL);
	if (hEnvPipe == INVALID_HANDLE_VALUE) return;
	
	hThread = CreateThread(NULL, 0, pipeReader, (LPVOID)hPipe, 0, NULL);
	if (hThread == NULL) return;
	
	/*vZXN*/
	si.cbSize = sizeof(SHELLEXECUTEINFO);
	si.fMask = SEE_MASK_NOCLOSEPROCESS;
	si.lpVerb = lpVerb;
	si.lpFile = sudo;
	si.lpParameters = lpParam;
	si.nShow = debug_mode ? SW_SHOWNORMAL : SW_HIDE;
	
	if (ShellExecuteEx(&si)) {
		HANDLE hEnvThread = CreateThread(NULL, 0, writeEnv, (LPVOID)hEnvPipe, 0, NULL);
		if (hEnvThread == NULL) return;
		
		WaitForSingleObject(si.hProcess, INFINITE);
		CloseHandle(si.hProcess);
		CloseHandle(hEnvThread);
	}
	
	CloseHandle(CreateFile(getPipeName(pipeName), GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL));
	CloseHandle(CreateFile(getEnvPipeName(pipeName), GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL));
	
	CloseHandle(hPipe);
	CloseHandle(hEnvPipe);
	CloseHandle(hCEvent);
	CloseHandle(hThread);
}
